# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys
import time
import xmlrpc.client

from lavacli import main
from lavacli.utils import safe_yaml


def test_jobs_cancel(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "cancel", "1234"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "scheduler.jobs.cancel", "args": ("1234",), "ret": None},
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_jobs_autocancel(setup, monkeypatch, capsys):
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "jobs", "autocancel", "--duration", "1440"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": "2023.10"},
            {
                "request": "scheduler.jobs.list",
                "args": ("RUNNING", None, 0, 100, None, False, 1440),
                "ret": [
                    {
                        "id": "16",
                        "description": "hello",
                        "device_type": "bbb",
                        "health": "Unknown",
                        "state": "Running",
                        "submitter": "lava",
                    },
                    {
                        "id": "17",
                        "description": "something",
                        "device_type": "qemu",
                        "health": "Unknown",
                        "state": "Running",
                        "submitter": "admin",
                    },
                ],
            },
            {
                "request": "scheduler.jobs.definition",
                "args": ("16",),
                "ret": "timeouts:\n  job:\n    minutes: 720\n",
            },
            {"request": "scheduler.jobs.cancel", "args": ("16",), "ret": None},
            {
                "request": "scheduler.jobs.definition",
                "args": ("17",),
                "ret": "timeouts:\n  job:\n    minutes: 1445\n",
            },
            {
                "request": "scheduler.jobs.list",
                "args": ("RUNNING", None, 0, 100, None, False, 1440),
                "ret": [],
            },
        ],
    )
    assert main() == 0
    assert (
        capsys.readouterr()[0]
        == """[CANCELED] 16: [lava] (hello) - bbb
[  SKIP  ] 17: [admin] (something) - qemu - timeout not reached yet\n"""
    )


def test_jobs_config(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "jobs", "config", "1234", "--dest", str(tmp_path)]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.configuration",
                "args": ("1234",),
                "ret": ["definition", "device", "dispatcher", "env", "env.dut"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec
    with (tmp_path / "definition.yaml").open() as f_in:
        assert f_in.read() == "definition"  # nosec
    with (tmp_path / "device.yaml").open() as f_in:
        assert f_in.read() == "device"  # nosec
    with (tmp_path / "dispatcher.yaml").open() as f_in:
        assert f_in.read() == "dispatcher"  # nosec
    with (tmp_path / "env.yaml").open() as f_in:
        assert f_in.read() == "env"  # nosec
    with (tmp_path / "env.dut.yaml").open() as f_in:
        assert f_in.read() == "env.dut"  # nosec


def test_jobs_definition(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "definition", "1234"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.definition",
                "args": ("1234",),
                "ret": "definition",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "definition\n"  # nosec


def test_jobs_list(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "list"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.list",
                "args": (None, None, 0, 25, 0, False),
                "ret": [
                    {
                        "id": "16",
                        "description": "hello",
                        "device_type": "bbb",
                        "health": "Complete",
                        "state": "Finished",
                        "submitter": "lava",
                    },
                    {
                        "id": "12",
                        "description": "something",
                        "device_type": "",
                        "health": "Unknown",
                        "state": "Running",
                        "submitter": "admin",
                    },
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Jobs (from 1 to 25):
* 16: Finished,Complete [lava] (hello) - bbb
* 12: Running,Unknown [admin] (something) - \n"""
    )


def test_jobs_list_since_verbose(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "jobs", "list", "--verbose", "--since", "5"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.list",
                "args": (None, None, 0, 25, 5, True),
                "ret": [
                    {
                        "id": "16",
                        "description": "hello",
                        "device_type": "bbb",
                        "health": "Complete",
                        "state": "Rinished",
                        "submitter": "lava",
                        "actual_device": "bbb-01",
                        "start_time": "12",
                        "end_time": "13",
                        "error_msg": None,
                        "error_type": None,
                    },
                    {
                        "id": "12",
                        "description": "something",
                        "device_type": "docker",
                        "health": "Unknown",
                        "state": "Running",
                        "submitter": "admin",
                        "actual_device": "docker-01",
                        "start_time": "45",
                        "end_time": "46",
                        "error_msg": "job error",
                        "error_type": "something is wrong",
                    },
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Jobs (from 1 to 25):
* 16: Rinished,Complete [lava] (hello) - bbb bbb-01 <12> <13>
* 12: Running,Unknown [admin] (something) - docker docker-01 <45> <46> something is wrong: job error\n"""
    )


def test_jobs_list_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "list", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.list",
                "args": (None, None, 0, 25, 0, False),
                "ret": [
                    {
                        "id": "16",
                        "description": "hello",
                        "device_type": "bbb",
                        "health": "Complete",
                        "state": "Rinished",
                        "submitter": "lava",
                    },
                    {
                        "id": "15",
                        "description": "world",
                        "device_type": "qemu",
                        "health": "Incomplete",
                        "state": "Finished",
                        "submitter": "lab",
                    },
                    {
                        "id": "14",
                        "description": "health",
                        "device_type": "docker",
                        "health": "Canceled",
                        "state": "Finished",
                        "submitter": "lava-health",
                    },
                    {
                        "id": "12",
                        "description": "something",
                        "device_type": "",
                        "health": "Unknown",
                        "state": "Running",
                        "submitter": "admin",
                    },
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == [  # nosec
        {
            "id": "16",
            "description": "hello",
            "device_type": "bbb",
            "health": "Complete",
            "state": "Rinished",
            "submitter": "lava",
        },
        {
            "id": "15",
            "description": "world",
            "device_type": "qemu",
            "health": "Incomplete",
            "state": "Finished",
            "submitter": "lab",
        },
        {
            "id": "14",
            "description": "health",
            "device_type": "docker",
            "health": "Canceled",
            "state": "Finished",
            "submitter": "lava-health",
        },
        {
            "id": "12",
            "description": "something",
            "device_type": "",
            "health": "Unknown",
            "state": "Running",
            "submitter": "admin",
        },
    ]


def test_jobs_list_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "list", "--yaml"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.list",
                "args": (None, None, 0, 25, 0, False),
                "ret": [
                    {
                        "id": "16",
                        "description": "hello",
                        "device_type": "bbb",
                        "health": "Complete",
                        "state": "Rinished",
                        "submitter": "lava",
                    },
                    {
                        "id": "15",
                        "description": "world",
                        "device_type": "qemu",
                        "health": "Incomplete",
                        "state": "Finished",
                        "submitter": "lab",
                    },
                    {
                        "id": "14",
                        "description": "health",
                        "device_type": "docker",
                        "health": "Canceled",
                        "state": "Finished",
                        "submitter": "lava-health",
                    },
                    {
                        "id": "12",
                        "description": "something",
                        "device_type": "",
                        "health": "Unknown",
                        "state": "Running",
                        "submitter": "admin",
                    },
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """- {description: hello, device_type: bbb, health: Complete, id: '16', state: Rinished,
  submitter: lava}
- {description: world, device_type: qemu, health: Incomplete, id: '15', state: Finished,
  submitter: lab}
- {description: health, device_type: docker, health: Canceled, id: '14', state: Finished,
  submitter: lava-health}
- {description: something, device_type: '', health: Unknown, id: '12', state: Running,
  submitter: admin}
"""
    )


def test_jobs_list_filtering(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "list", "--state", "RUNNING"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.list",
                "args": ("RUNNING", None, 0, 25, 0, False),
                "ret": [],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Jobs (from 1 to 25):\n"  # nosec


def test_jobs_list_filtering2(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "jobs",
            "list",
            "--health",
            "CANCELED",
            "--start",
            "45",
            "--limit",
            "56",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.list",
                "args": (None, "CANCELED", 45, 56, 0, False),
                "ret": [],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Jobs (from 46 to 101):\n"  # nosec


# TODO: test with 2018.6 and also --start/--end
def test_jobs_logs(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "logs", "1234"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 0, None),
                "ret": (
                    False,
                    '- {"dt": "2018-04-23T12:07:02.569264", "lvl": "info", "msg": "lava-dispatcher, installed at version: 2019.1-1"}',
                ),
            },
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 1, None),
                "ret": (
                    True,
                    """- {"dt": "2018-04-23T12:07:02.572789", "lvl": "results", "msg": {"case": "validate", "definition": "lava", "result": "pass"}}
- {"dt": "2018-04-23T12:07:02.573414", "lvl": "debug", "msg": "start: 1.1 download-retry (timeout 00:02:00) [common]"}""",
                ),
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("1234",),
                "ret": {
                    "id": "1234",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    lines = capsys.readouterr()[0].split("\n")
    assert (  # nosec
        lines[0]
        == "2018-04-23T12:07:02 lava-dispatcher, installed at version: 2019.1-1"
    )
    assert lines[1][:20] == "2018-04-23T12:07:02 "  # nosec
    assert safe_yaml.load(lines[1][20:]) == {  # nosec
        "case": "validate",
        "definition": "lava",
        "result": "pass",
    }
    assert (  # nosec
        lines[2]
        == "2018-04-23T12:07:02 start: 1.1 download-retry (timeout 00:02:00) [common]"
    )


def test_jobs_logs_failure_comment_and_polling(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 10  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "jobs", "logs", "--polling", "10", "1234"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 0, None),
                "ret": (
                    False,
                    '- {"dt": "2018-04-23T12:07:02.569264", "lvl": "info", "msg": "lava-dispatcher, installed at version: 2019.1-1"}',
                ),
            },
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 1, None),
                "ret": (
                    True,
                    """- {"dt": "2018-04-23T12:07:02.572789", "lvl": "results", "msg": {"case": "validate", "definition": "lava", "result": "pass"}}
- {"dt": "2018-04-23T12:07:02.573414", "lvl": "debug", "msg": "start: 1.1 download-retry (timeout 00:02:00) [common]"}""",
                ),
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("1234",),
                "ret": {
                    "id": "1234",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Incomplete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "A small issue was found",
                },
            },
        ],
    )
    assert main() == 0  # nosec
    lines = capsys.readouterr()[0].split("\n")
    assert (  # nosec
        lines[0]
        == "2018-04-23T12:07:02 lava-dispatcher, installed at version: 2019.1-1"
    )
    assert lines[1][:20] == "2018-04-23T12:07:02 "  # nosec
    assert safe_yaml.load(lines[1][20:]) == {  # nosec
        "case": "validate",
        "definition": "lava",
        "result": "pass",
    }
    assert (  # nosec
        lines[2]
        == "2018-04-23T12:07:02 start: 1.1 download-retry (timeout 00:02:00) [common]"
    )
    assert lines[3].endswith(  # nosec
        "lavacli] Failure comment: A small issue was found"
    )


def test_jobs_logs_filtering(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "jobs", "logs", "1234", "--filters", "info,debug"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 0, None),
                "ret": (
                    False,
                    '- {"dt": "2018-04-23T12:07:02.569264", "lvl": "info", "msg": "lava-dispatcher, installed at version: 2019.1-1"}',
                ),
            },
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 1, None),
                "ret": (
                    True,
                    """- {"dt": "2018-04-23T12:07:02.572789", "lvl": "results", "msg": {"case": "validate", "definition": "lava", "result": "pass"}}
- {"dt": "2018-04-23T12:07:02.573414", "lvl": "debug", "msg": "start: 1.1 download-retry (timeout 00:02:00) [common]"}""",
                ),
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("1234",),
                "ret": {
                    "id": "1234",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """2018-04-23T12:07:02 lava-dispatcher, installed at version: 2019.1-1
2018-04-23T12:07:02 start: 1.1 download-retry (timeout 00:02:00) [common]
"""
    )


def test_jobs_logs_raw(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "logs", "1234", "--raw"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 0, None),
                "ret": (
                    False,
                    '- {"dt": "2018-04-23T12:07:02.569264", "lvl": "info", "msg": "lava-dispatcher, installed at version: 2019.1-1"}',
                ),
            },
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 1, None),
                "ret": (
                    True,
                    """- {"dt": "2018-04-23T12:07:02.572789", "lvl": "results", "msg": {"case": "validate", "definition": "lava", "result": "pass"}}
- {"dt": "2018-04-23T12:07:02.573414", "lvl": "debug", "msg": "start: 1.1 download-retry (timeout 00:02:00) [common]"}""",
                ),
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("1234",),
                "ret": {
                    "id": "1234",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """- {"dt": "2018-04-23T12:07:02.569264", "lvl": "info", "msg": "lava-dispatcher, installed at version: 2019.1-1"}
- {"dt": "2018-04-23T12:07:02.572789", "lvl": "results", "msg": {"case": "validate", "definition": "lava", "result": "pass"}}
- {"dt": "2018-04-23T12:07:02.573414", "lvl": "debug", "msg": "start: 1.1 download-retry (timeout 00:02:00) [common]"}
"""
    )


def test_jobs_logs_raw_filter(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "jobs", "logs", "1234", "--raw", "--filters", "debug,results"],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 0, None),
                "ret": (
                    False,
                    '- {"dt": "2018-04-23T12:07:02.569264", "lvl": "info", "msg": "lava-dispatcher, installed at version: 2019.1-1"}',
                ),
            },
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 1, None),
                "ret": (
                    True,
                    """- {"dt": "2018-04-23T12:07:02.572789", "lvl": "results", "msg": {"case": "validate", "definition": "lava", "result": "pass"}}
- {"dt": "2018-04-23T12:07:02.573414", "lvl": "debug", "msg": "start: 1.1 download-retry (timeout 00:02:00) [common]"}""",
                ),
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("1234",),
                "ret": {
                    "id": "1234",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """- {"dt": "2018-04-23T12:07:02.572789", "lvl": "results", "msg": {"case": "validate", "definition": "lava", "result": "pass"}}
- {"dt": "2018-04-23T12:07:02.573414", "lvl": "debug", "msg": "start: 1.1 download-retry (timeout 00:02:00) [common]"}
"""
    )


def test_jobs_queue(setup, monkeypatch, capsys):
    version = "2019.01"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "queue"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.queue",
                "args": (None, 0, 25),
                "ret": [
                    {
                        "id": 12,
                        "description": None,
                        "requested_device_type": "qemu",
                        "submitter": "lava-health",
                    },
                    {
                        "id": 14,
                        "description": "qemu health",
                        "requested_device_type": "qemu",
                        "submitter": "lava-health",
                    },
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Jobs (from 1 to 25):
* 12: lava-health () - qemu
* 14: lava-health (qemu health) - qemu
"""
    )


def test_jobs_queue_json(setup, monkeypatch, capsys):
    version = "2019.01"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "queue", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.queue",
                "args": (None, 0, 25),
                "ret": [
                    {
                        "id": 12,
                        "description": None,
                        "requested_device_type": "qemu",
                        "submitter": "lava-health",
                    },
                    {
                        "id": 14,
                        "description": "qemu health",
                        "requested_device_type": "qemu",
                        "submitter": "lava-health",
                    },
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == [  # nosec
        {
            "id": 12,
            "description": None,
            "requested_device_type": "qemu",
            "submitter": "lava-health",
        },
        {
            "id": 14,
            "description": "qemu health",
            "requested_device_type": "qemu",
            "submitter": "lava-health",
        },
    ]


def test_jobs_queue_yaml(setup, monkeypatch, capsys):
    version = "2019.01"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "queue", "--yaml"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.queue",
                "args": (None, 0, 25),
                "ret": [
                    {
                        "id": 12,
                        "description": None,
                        "requested_device_type": "qemu",
                        "submitter": "lava-health",
                    },
                    {
                        "id": 14,
                        "description": "qemu health",
                        "requested_device_type": "qemu",
                        "submitter": "lava-health",
                    },
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """- {description: null, id: 12, requested_device_type: qemu, submitter: lava-health}
- {description: qemu health, id: 14, requested_device_type: qemu, submitter: lava-health}
"""
    )


def test_jobs_queue_filtering(setup, monkeypatch, capsys):
    version = "2019.01"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "queue", "qemu"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "scheduler.jobs.queue", "args": (["qemu"], 0, 25), "ret": []},
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Jobs (from 1 to 25):\n"  # nosec

    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "queue", "qemu", "panda"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.queue",
                "args": (["qemu", "panda"], 0, 25),
                "ret": [],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Jobs (from 1 to 25):\n"  # nosec

    version = "2019.01"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "jobs", "queue", "--start", "1", "--limit", "3", "qemu"],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "scheduler.jobs.queue", "args": (["qemu"], 1, 3), "ret": []},
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Jobs (from 2 to 4):\n"  # nosec


def test_jobs_resubmit(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "resubmit", "1234"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "scheduler.jobs.resubmit", "args": ("1234",), "ret": 1234},
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "1234\n"  # nosec


def test_jobs_resubmit_url(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "--uri",
            "https://localhost:8000/RPC2",
            "jobs",
            "resubmit",
            "1234",
            "--url",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "scheduler.jobs.resubmit", "args": ("1234",), "ret": 1234},
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "https://localhost:8000/scheduler/job/1234\n"
    )


def test_jobs_resubmit_mutlinode(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "resubmit", "1234"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.resubmit",
                "args": ("1234",),
                "ret": [1234, 1235],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "1234\n1235\n"  # nosec


def test_jobs_resubmit_mutlinode_url(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "--uri",
            "https://example.com/RPC2",
            "jobs",
            "resubmit",
            "1234",
            "--url",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.resubmit",
                "args": ("1234",),
                "ret": [1234, 1345],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == "https://example.com/scheduler/job/1234\nhttps://example.com/scheduler/job/1345\n"
    )


def test_jobs_resubmit_follow(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "jobs", "resubmit", "1234", "--follow", "--filters", "info,debug"],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "scheduler.jobs.resubmit", "args": ("1234",), "ret": 1234},
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 0, None),
                "ret": (
                    False,
                    '- {"dt": "2018-04-23T12:07:02.569264", "lvl": "info", "msg": "lava-dispatcher, installed at version: 2019.1-1"}',
                ),
            },
            {
                "request": "scheduler.jobs.logs",
                "args": ("1234", 1, None),
                "ret": (
                    True,
                    """- {"dt": "2018-04-23T12:07:02.572789", "lvl": "results", "msg": {"case": "validate", "definition": "lava", "result": "pass"}}
- {"dt": "2018-04-23T12:07:02.573414", "lvl": "debug", "msg": "start: 1.1 download-retry (timeout 00:02:00) [common]"}""",
                ),
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("1234",),
                "ret": {
                    "id": "1234",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    lines = capsys.readouterr()[0].split("\n")
    assert lines[0].endswith("[lavacli] Job 1234 submitted")  # nosec
    assert (  # nosec
        lines[1]
        == "2018-04-23T12:07:02 lava-dispatcher, installed at version: 2019.1-1"
    )
    assert (  # nosec
        lines[2]
        == "2018-04-23T12:07:02 start: 1.1 download-retry (timeout 00:02:00) [common]"
    )


def test_jobs_run(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    with (tmp_path / "job.yaml").open("w") as f_out:
        f_out.write("job definition")
    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "jobs", "run", str(tmp_path / "job.yaml")]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.submit",
                "args": ("job definition",),
                "ret": "4567",
            },
            {
                "request": "scheduler.jobs.logs",
                "args": ("4567", 0, None),
                "ret": (
                    False,
                    '- {"dt": "2018-04-23T12:07:02.569264", "lvl": "info", "msg": "lava-dispatcher, installed at version: 2019.1-1"}',
                ),
            },
            {
                "request": "scheduler.jobs.logs",
                "args": ("4567", 1, None),
                "ret": (
                    True,
                    """- {"dt": "2018-04-23T12:07:02.572789", "lvl": "results", "msg": {"case": "validate", "definition": "lava", "result": "pass"}}
- {"dt": "2018-04-23T12:07:02.573414", "lvl": "debug", "msg": "start: 1.1 download-retry (timeout 00:02:00) [common]"}""",
                ),
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("4567",),
                "ret": {
                    "id": "4567",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    lines = capsys.readouterr()[0].split("\n")
    assert lines[0].endswith("[lavacli] Job 4567 submitted")  # nosec
    assert (  # nosec
        lines[1]
        == "2018-04-23T12:07:02 lava-dispatcher, installed at version: 2019.1-1"
    )
    assert lines[2][:20] == "2018-04-23T12:07:02 "  # nosec
    assert safe_yaml.load(lines[2][20:]) == {  # nosec
        "case": "validate",
        "definition": "lava",
        "result": "pass",
    }
    assert (  # nosec
        lines[3]
        == "2018-04-23T12:07:02 start: 1.1 download-retry (timeout 00:02:00) [common]"
    )


def test_jobs_show(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "show", "789"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.show",
                "args": ("789",),
                "ret": {
                    "id": "789",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """id          : 789
description : desc
submitter   : lava-admin
device-type : qemu
device      : qemu01
health-check: False
state       : Finished
Health      : Complete
pipeline    : True
tags        : []
visibility  : Publicly visible
submit time : 20180128T01:01:01
start time  : 20180128T01:01:01
end time    : 20180128T01:01:01
"""
    )


def test_jobs_show_failure_comment(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "show", "789"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.show",
                "args": ("789",),
                "ret": {
                    "id": "789",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Incomplete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "Something went wrong",
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """id          : 789
description : desc
submitter   : lava-admin
device-type : qemu
device      : qemu01
health-check: False
state       : Finished
Health      : Incomplete
failure     : Something went wrong
pipeline    : True
tags        : []
visibility  : Publicly visible
submit time : 20180128T01:01:01
start time  : 20180128T01:01:01
end time    : 20180128T01:01:01
"""
    )


def test_jobs_show_json(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "show", "789", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.show",
                "args": ("789",),
                "ret": {
                    "id": "789",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == {  # nosec
        "id": "789",
        "description": "desc",
        "device": "qemu01",
        "device_type": "qemu",
        "health_check": False,
        "pipeline": True,
        "health": "Complete",
        "state": "Finished",
        "submitter": "lava-admin",
        "submit_time": "20180128T01:01:01",
        "start_time": "20180128T01:01:01",
        "end_time": "20180128T01:01:01",
        "tags": [],
        "visibility": "Publicly visible",
        "failure_comment": None,
    }


def test_jobs_show_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "show", "789", "--yaml"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.show",
                "args": ("789",),
                "ret": {
                    "id": "789",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """description: desc
device: qemu01
device_type: qemu
end_time: 20180128T01:01:01
failure_comment: null
health: Complete
health_check: false
id: '789'
pipeline: true
start_time: 20180128T01:01:01
state: Finished
submit_time: 20180128T01:01:01
submitter: lava-admin
tags: []
visibility: Publicly visible
"""
    )


def test_jobs_submit(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "job.yaml").open("w") as f_out:
        f_out.write("job definition as yaml")
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "jobs", "submit", str(tmp_path / "job.yaml")]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.submit",
                "args": ("job definition as yaml",),
                "ret": 5689,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "5689\n"  # nosec


def test_jobs_submit_url(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "job.yaml").open("w") as f_out:
        f_out.write("job definition as yaml")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "--uri",
            "https://localhost:8000/RPC2",
            "jobs",
            "submit",
            str(tmp_path / "job.yaml"),
            "--url",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.submit",
                "args": ("job definition as yaml",),
                "ret": 5689,
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "https://localhost:8000/scheduler/job/5689\n"
    )


def test_jobs_submit_url_identity_default(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "job.yaml").open("w") as f_out:
        f_out.write("job definition as yaml")
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "jobs", "submit", str(tmp_path / "job.yaml"), "--url"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.submit",
                "args": ("job definition as yaml",),
                "ret": 5689,
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "https://lava.example.com/scheduler/job/5689\n"
    )


def test_jobs_submit_url_identity_admin(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "job.yaml").open("w") as f_out:
        f_out.write("job definition as yaml")
    with (tmp_path / "lavacli.yaml").open("w") as f_conf:
        safe_yaml.dump(
            {
                "default": {"uri": "https://lava.example.com/RPC2"},
                "admin": {"uri": "https://localhost:8001/RPC2"},
            },
            f_conf,
        )
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "-i",
            "admin",
            "jobs",
            "submit",
            str(tmp_path / "job.yaml"),
            "--url",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.submit",
                "args": ("job definition as yaml",),
                "ret": 5689,
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "https://localhost:8001/scheduler/job/5689\n"
    )


def test_jobs_submit_multiple(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "job1.yaml").open("w") as f_out:
        f_out.write("first job definition as yaml")
    with (tmp_path / "job2.yaml").open("w") as f_out:
        f_out.write("second job definition as yaml")
    with (tmp_path / "job3.yaml").open("w") as f_out:
        f_out.write("third job definition as yaml")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "jobs",
            "submit",
            str(tmp_path / "job1.yaml"),
            str(tmp_path / "job2.yaml"),
            str(tmp_path / "job3.yaml"),
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.submit",
                "args": ("first job definition as yaml",),
                "ret": 5689,
            },
            {
                "request": "scheduler.jobs.submit",
                "args": ("second job definition as yaml",),
                "ret": 5690,
            },
            {
                "request": "scheduler.jobs.submit",
                "args": ("third job definition as yaml",),
                "ret": [5691, 5692],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "5689\n5690\n5691\n5692\n"  # nosec


def test_jobs_submit_multinode(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "job.yaml").open("w") as f_out:
        f_out.write("job definition as yaml")
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "jobs", "submit", str(tmp_path / "job.yaml")]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.submit",
                "args": ("job definition as yaml",),
                "ret": [5689, 5698],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "5689\n5698\n"  # nosec


def test_jobs_submit_multinode_url(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "job.yaml").open("w") as f_out:
        f_out.write("job definition as yaml")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "--uri",
            "https://localhost:8000/RPC2",
            "jobs",
            "submit",
            str(tmp_path / "job.yaml"),
            "--url",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.submit",
                "args": ("job definition as yaml",),
                "ret": [5689, 5698],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == "https://localhost:8000/scheduler/job/5689\nhttps://localhost:8000/scheduler/job/5698\n"
    )


def test_jobs_wait(setup, monkeypatch, capsys):
    version = "2019.1"
    now = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(sys, "argv", ["lavacli", "jobs", "wait", "1234"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.show",
                "args": ("1234",),
                "ret": {
                    "id": "1234",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Unknown",
                    "state": "Submitted",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": None,
                    "end_time": None,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("1234",),
                "ret": {
                    "id": "1234",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Unknown",
                    "state": "Running",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": None,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("1234",),
                "ret": {
                    "id": "1234",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Unknown",
                    "state": "Running",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": None,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("1234",),
                "ret": {
                    "id": "1234",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )

    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Submitted\nRunning.\n"  # nosec


def test_jobs_submit_follow(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "job.yaml").open("w") as f_out:
        f_out.write("job definition as yaml")
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "jobs", "submit", "--follow", str(tmp_path / "job.yaml")],
    )
    now = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.jobs.submit",
                "args": ("job definition as yaml",),
                "ret": 5689,
            },
            {
                "request": "scheduler.jobs.logs",
                "args": ("5689", 0, None),
                "ret": (
                    True,
                    '- {"dt": "2018-04-23T12:07:02.569264", "lvl": "info", "msg": "lava-dispatcher, installed at version: 2019.1-1"}',
                ),
            },
            {
                "request": "scheduler.jobs.show",
                "args": ("5689",),
                "ret": {
                    "id": "5689",
                    "description": "desc",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-admin",
                    "submit_time": now,
                    "start_time": now,
                    "end_time": now,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": None,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert "lava-dispatcher, installed at version: 2019.1" in capsys.readouterr()[0]



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import argparse
import json
import sys

from lavacli.utils import safe_yaml


def configure_parser(parser, version):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    # "add"
    dt_add = sub.add_parser("add", help="add a device type")
    dt_add.add_argument("name", help="name of the device-type")
    dt_add.add_argument("--description", default=None, help="device-type description")
    dt_add.add_argument(
        "--hide",
        dest="display",
        action="store_false",
        default=True,
        help="device is hidden in the UI",
    )
    dt_add.add_argument(
        "--owners-only",
        action="store_true",
        default=False,
        help="devices are only visible to owners",
    )
    dt_health = dt_add.add_argument_group("health check")
    dt_health.add_argument(
        "--health-frequency",
        default=24,
        type=int,
        help="how often to run health checks.",
    )
    dt_health.add_argument(
        "--health-denominator",
        default="hours",
        choices=["hours", "jobs"],
        help="initiate health checks by hours or by jobs.",
    )

    # "aliases"
    dt_aliases = sub.add_parser(
        "aliases", help="manage aliases for the given device-type"
    )
    aliases_sub = dt_aliases.add_subparsers(
        dest="sub_sub_sub_command", help="Sub commands"
    )
    aliases_sub.required = True

    aliases_add = aliases_sub.add_parser("add", help="add aliases")
    aliases_add.add_argument("name", help="name of the device-type")
    aliases_add.add_argument("alias", help="name of alias")

    aliases_delete = aliases_sub.add_parser("delete", help="delete aliases")
    aliases_delete.add_argument("name", help="name of the device-type")
    aliases_delete.add_argument("alias", help="name of alias")

    aliases_list = aliases_sub.add_parser(
        "list", help="list aliases for the device-type"
    )
    aliases_list.add_argument("name", help="device-type")
    out_format = aliases_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        default=None,
        action="store_const",
        const="yaml",
        help="print as yaml",
    )

    # "heath-check"
    dt_hc = sub.add_parser("health-check", help="device-type health-check")
    dt_sub = dt_hc.add_subparsers(dest="sub_sub_sub_command", help="Sub commands")
    dt_sub.required = True
    if version >= (2022, 4):
        dt_delete = dt_sub.add_parser(
            "delete", help="delete the device-type health-check"
        )
        dt_delete.add_argument("name", help="name of the device-type")

    dt_get = dt_sub.add_parser("get", help="get the device-type health-check")
    dt_get.add_argument("name", help="name of the device-type")

    dt_set = dt_sub.add_parser("set", help="set the device-type health-check")
    dt_set.add_argument("name", help="name of the device-type")
    dt_set.add_argument(
        "definition", type=argparse.FileType("r"), help="health-check definition"
    )

    # "list"
    dt_list = sub.add_parser("list", help="list available device-types")
    dt_list.add_argument(
        "--all",
        "-a",
        dest="show_all",
        default=False,
        action="store_true",
        help="show all device types in the database, " "including non-installed ones",
    )
    out_format = dt_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        default=None,
        action="store_const",
        const="yaml",
        help="print as yaml",
    )

    # "perms"
    if version >= (2023, 3):
        dt_perms = sub.add_parser("perms", help="permissions")
        perms_sub = dt_perms.add_subparsers(
            dest="sub_sub_sub_command", help="Sub commands"
        )
        perms_sub.required = True

        perms_add = perms_sub.add_parser("add", help="add permissions")
        perms_add.add_argument("name", help="name of the device-type")
        perms_add.add_argument("group", help="group")
        perms_add.add_argument("permission", help="permission")

        perms_delete = perms_sub.add_parser("delete", help="delete permissions")
        perms_delete.add_argument("name", help="name of the device-type")
        perms_delete.add_argument("group", help="group")
        perms_delete.add_argument("permission", help="permission")

        perms_list = perms_sub.add_parser("list", help="list permissions")
        perms_list.add_argument("name", help="name of the device-type")
        out_format = perms_list.add_mutually_exclusive_group()
        out_format.add_argument(
            "--json",
            dest="output_format",
            default=None,
            action="store_const",
            const="json",
            help="print as json",
        )
        out_format.add_argument(
            "--yaml",
            dest="output_format",
            default=None,
            action="store_const",
            const="yaml",
            help="print as yaml",
        )

    # "show"
    dt_show = sub.add_parser("show", help="show device-type details")
    dt_show.add_argument("name", help="name of the device-type")
    out_format = dt_show.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        default=None,
        help="print as yaml",
    )

    # "template"
    dt_template = sub.add_parser("template", help="device-type template")
    dt_sub = dt_template.add_subparsers(dest="sub_sub_sub_command", help="Sub commands")
    dt_sub.required = True
    if version >= (2022, 4):
        dt_delete = dt_sub.add_parser(
            "delete", help="delete the custom device-type template"
        )
        dt_delete.add_argument("name", help="name of the device-type")

    dt_get = dt_sub.add_parser("get", help="get the device-type template")
    dt_get.add_argument("name", help="name of the device-type")

    dt_set = dt_sub.add_parser("set", help="set the device-type template")
    dt_set.add_argument("name", help="name of the device-type")
    dt_set.add_argument("template", type=argparse.FileType("r"), help="template file")

    # "update"
    dt_update = sub.add_parser("update", help="update device-type")
    dt_update.add_argument("name", help="name of the device-type")
    dt_update.add_argument(
        "--description", default=None, help="device-type description"
    )

    visibility = dt_update.add_mutually_exclusive_group()
    visibility.add_argument(
        "--hide",
        dest="display",
        action="store_false",
        default=None,
        help="device-type is hidden in the UI",
    )
    visibility.add_argument(
        "--show",
        dest="display",
        action="store_true",
        help="device-type is visible in the UI",
    )

    owner = dt_update.add_mutually_exclusive_group()
    owner.add_argument(
        "--owners-only",
        action="store_true",
        dest="owners_only",
        default=None,
        help="devices are only visible to owners",
    )
    owner.add_argument(
        "--public",
        action="store_false",
        dest="owners_only",
        help="devices are visible to all users",
    )

    dt_health = dt_update.add_argument_group("health check")
    dt_health.add_argument(
        "--health-frequency",
        default=None,
        type=int,
        help="how often to run health checks.",
    )
    dt_health.add_argument(
        "--health-denominator",
        default=None,
        choices=["hours", "jobs"],
        help="initiate health checks by hours or by jobs.",
    )

    health = dt_health.add_mutually_exclusive_group()
    health.add_argument(
        "--health-disabled",
        default=None,
        action="store_true",
        help="disable health checks",
    )
    health.add_argument(
        "--health-active",
        dest="health_disabled",
        action="store_false",
        help="activate health checks",
    )


def help_string():
    return "manage device-types"


def handle_add(proxy, options, _):
    proxy.scheduler.device_types.add(
        options.name,
        options.description,
        options.display,
        options.owners_only,
        options.health_frequency,
        options.health_denominator,
    )
    return 0


def handle_aliases(proxy, options, _):
    if options.sub_sub_sub_command == "add":
        proxy.scheduler.device_types.aliases.add(options.name, options.alias)
    elif options.sub_sub_sub_command == "list":
        aliases = proxy.scheduler.device_types.aliases.list(options.name)
        if options.output_format == "json":
            print(json.dumps(aliases))
        elif options.output_format == "yaml":
            safe_yaml.dump(aliases, sys.stdout)
        else:
            print("Aliases:")
            for alias in aliases:
                print("* %s" % alias)
    else:
        assert options.sub_sub_sub_command == "delete"
        proxy.scheduler.device_types.aliases.delete(options.name, options.alias)
    return 0


def handle_hc(proxy, options, _):
    if options.sub_sub_sub_command == "delete":
        template = proxy.scheduler.device_types.set_health_check(options.name, "")
    elif options.sub_sub_sub_command == "get":
        template = proxy.scheduler.device_types.get_health_check(options.name)
        print(str(template).rstrip("\n"))
    else:
        hc = options.definition.read()
        proxy.scheduler.device_types.set_health_check(options.name, hc)
    return 0


def handle_list(proxy, options, _):
    device_types = proxy.scheduler.device_types.list(options.show_all)

    if options.output_format == "json":
        print(json.dumps(device_types))
    elif options.output_format == "yaml":
        safe_yaml.dump(device_types, sys.stdout)
    else:
        print("Device-Types:")
        for dt in device_types:
            print(f"* {dt['name']} ({dt['devices']})")
    return 0


def handle_perms(proxy, options, _):
    if options.sub_sub_sub_command == "add":
        proxy.scheduler.device_types.perms_add(
            options.name, options.group, options.permission
        )
    elif options.sub_sub_sub_command == "delete":
        proxy.scheduler.device_types.perms_delete(
            options.name, options.group, options.permission
        )
    elif options.sub_sub_sub_command == "list":
        perms = proxy.scheduler.device_types.perms_list(options.name)
        if options.output_format == "json":
            print(json.dumps(perms))
        elif options.output_format == "yaml":
            safe_yaml.dump(perms, sys.stdout)
        else:
            print("Permissions:")
            for perm in perms:
                print("* %s %s" % (perm["group"], perm["name"]))
    return 0


def handle_show(proxy, options, config):
    dt = proxy.scheduler.device_types.show(options.name)

    if options.output_format == "json":
        print(json.dumps(dt))
    elif options.output_format == "yaml":
        safe_yaml.dump(dt, sys.stdout)
    else:
        print("name            : %s" % dt["name"])
        print("description     : %s" % dt["description"])
        print("display         : %s" % dt["display"])
        if config["version"] <= (2019, 7):
            print("owners only     : %s" % dt["owners_only"])
        print("health disabled : %s" % dt["health_disabled"])
        print("aliases         : %s" % dt["aliases"])
        print("devices         : %s" % dt["devices"])
    return 0


def handle_template(proxy, options, _):
    if options.sub_sub_sub_command == "delete":
        template = proxy.scheduler.device_types.set_template(options.name, "")
    elif options.sub_sub_sub_command == "get":
        template = proxy.scheduler.device_types.get_template(options.name)
        print(str(template).rstrip("\n"))
    else:
        template = options.template.read()
        proxy.scheduler.device_types.set_template(options.name, template)
    return 0


def handle_update(proxy, options, _):
    proxy.scheduler.device_types.update(
        options.name,
        options.description,
        options.display,
        options.owners_only,
        options.health_frequency,
        options.health_denominator,
        options.health_disabled,
    )
    return 0


def handle(proxy, options, config):
    handlers = {
        "add": handle_add,
        "aliases": handle_aliases,
        "health-check": handle_hc,
        "list": handle_list,
        "perms": handle_perms,
        "show": handle_show,
        "template": handle_template,
        "update": handle_update,
    }
    return handlers[options.sub_sub_command](proxy, options, config)



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys

from lavacli.utils import safe_yaml


def configure_parser(parser, version):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    # "add"
    aliases_add = sub.add_parser("add", help="add an alias")
    aliases_add.add_argument("alias", help="alias name")
    if version >= (2019, 5):
        aliases_add.add_argument("device_type", help="device-type")

    # "delete"
    aliases_del = sub.add_parser("delete", help="delete an alias")
    aliases_del.add_argument("alias", help="alias name")

    # "list"
    aliases_list = sub.add_parser("list", help="list available aliases")
    out_format = aliases_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        default=None,
        action="store_const",
        const="yaml",
        help="print as yaml",
    )

    # "show"
    aliases_show = sub.add_parser("show", help="show alias details")
    aliases_show.add_argument("alias", help="alias")
    out_format = aliases_show.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        default=None,
        help="print as yaml",
    )


def help_string():
    return "manage device-type aliases"


def handle_add(proxy, options, config):
    if config["version"] >= (2019, 5):
        proxy.scheduler.aliases.add(options.alias, options.device_type)
    else:
        proxy.scheduler.aliases.add(options.alias)
    return 0


def handle_delete(proxy, options, _):
    proxy.scheduler.aliases.delete(options.alias)
    return 0


def handle_list(proxy, options, _):
    aliases = proxy.scheduler.aliases.list()
    if options.output_format == "json":
        print(json.dumps(aliases))
    elif options.output_format == "yaml":
        safe_yaml.dump(aliases, sys.stdout)
    else:
        print("Aliases:")
        for alias in aliases:
            print("* %s" % alias)
    return 0


def handle_show(proxy, options, config):
    alias = proxy.scheduler.aliases.show(options.alias)
    if options.output_format == "json":
        print(json.dumps(alias))
    elif options.output_format == "yaml":
        safe_yaml.dump(alias, sys.stdout)
    else:
        if config["version"] >= (2019, 5):
            print("name       : %s" % alias["name"])
            print("device-type: %s" % alias["device_type"])
        else:
            print("name        : %s" % alias["name"])
            print("device-types:")
            for dt in alias["device_types"]:
                print("* %s" % dt)
    return 0


def handle(proxy, options, config):
    handlers = {
        "add": handle_add,
        "delete": handle_delete,
        "list": handle_list,
        "show": handle_show,
    }
    return handlers[options.sub_sub_command](proxy, options, config)



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import contextlib
import pathlib
import shutil
import time
import xmlrpc.client

from lavacli.utils import flow_yaml


def configure_parser(parser, version):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    # "active"
    sub.add_parser("active", help="activate the system")

    #  "api"
    sub.add_parser("api", help="print server API version")

    # "certificate"
    if version >= (2020, 4):
        sub.add_parser("certificate")

    # "export"
    sys_export = sub.add_parser("export", help="export server configuration")
    sys_export.add_argument("name", help="name of the export")
    sys_export.add_argument(
        "--full",
        default=False,
        action="store_true",
        help="do a full export, including retired devices",
    )

    # "import"
    # sub.add_parser("import", help="import server configuration")

    # "maintenance"
    sys_maintenance = sub.add_parser("maintenance", help="maintenance the system")
    sys_maintenance.add_argument(
        "--force",
        default=False,
        action="store_true",
        help="force maintenance by canceling jobs",
    )
    sys_maintenance.add_argument(
        "--exclude",
        default=None,
        help="comma separated list of workers to keep untouched",
    )

    # "methods"
    sys_methods = sub.add_parser("methods", help="list methods")
    sys_sub = sys_methods.add_subparsers(
        dest="sub_sub_sub_command", help="Sub commands"
    )
    sys_sub.required = True
    sys_sub.add_parser("list", help="list available methods")

    sys_help = sys_sub.add_parser("help", help="method help")
    sys_help.add_argument("method", help="method name")

    sys_signature = sys_sub.add_parser("signature", help="method signature")
    sys_signature.add_argument("method", help="method name")

    # "version"
    sub.add_parser("version", help="print the server version")

    # "whoami"
    sub.add_parser("whoami", help="print the current username")


def help_string():
    return "system information"


def handle_active(proxy, _, __):
    print("Activate workers:")
    workers = proxy.scheduler.workers.list()
    for worker in workers:
        print("* %s" % worker)
        proxy.scheduler.workers.update(worker, None, "ACTIVE")
    return 0


def handle_api(proxy, _, __):
    print(proxy.system.api_version())
    return 0


def handle_export(proxy, options, config):
    print("Export to %s" % options.name)
    dest = pathlib.Path(options.name)
    with contextlib.suppress(FileNotFoundError):
        shutil.rmtree(str(dest))
    with contextlib.suppress(FileExistsError):
        dest.mkdir(mode=0o755)

    print("Listing aliases")
    aliases = []
    for alias in proxy.scheduler.aliases.list():
        print("* %s" % alias)
        aliases.append(proxy.scheduler.aliases.show(alias))

    print("Listing tags")
    tags = []
    for tag in proxy.scheduler.tags.list():
        print("* %s" % tag["name"])
        tags.append(tag)

    print("Listing workers")
    workers = []
    for worker in proxy.scheduler.workers.list():
        print("* %s" % worker)
        w = proxy.scheduler.workers.show(worker)
        workers.append(
            {
                "hostname": w["hostname"],
                "description": w["description"],
                "state": w["state"],
                "health": w["health"],
            }
        )

    print("Listing device-types")
    (dest / "device-types").mkdir(mode=0o755)
    device_types = []
    for device_type in proxy.scheduler.device_types.list():
        print("* %s" % device_type["name"])
        dt = proxy.scheduler.device_types.show(device_type["name"])
        device_types.append(
            {
                "name": dt["name"],
                "description": dt["description"],
                "display": dt["display"],
                "health_disabled": dt["health_disabled"],
                "owners_only": dt["owners_only"],
                "aliases": dt["aliases"],
            }
        )

        try:
            dt_template = proxy.scheduler.device_types.get_template(dt["name"])
        except xmlrpc.client.Fault as exc:
            if exc.faultCode == 404:
                print("  => No template found")
                continue
            raise
        with (dest / "device-types" / ("%s.jinja2" % dt["name"])).open(
            "w", encoding="utf-8"
        ) as f_out:
            f_out.write(str(dt_template))

    print("Listing devices")
    (dest / "devices").mkdir(mode=0o755)
    devices = []
    for device in proxy.scheduler.devices.list(options.full):
        print("* %s" % device["hostname"])
        d = proxy.scheduler.devices.show(device["hostname"])
        devices.append(
            {
                "hostname": d["hostname"],
                "description": d["description"],
                "device_type": d["device_type"],
                "pipeline": d["pipeline"],
                "worker": d["worker"],
                "state": d["state"],
                "health": d["health"],
                "public": d["public"],
                "user": d["user"],
                "group": d["group"],
                "tags": d["tags"],
            }
        )

        try:
            device_dict = proxy.scheduler.devices.get_dictionary(device["hostname"])
        except xmlrpc.client.Fault as exc:
            if exc.faultCode == 404:
                print("  => No device dict found")
                continue
            raise
        with (dest / "devices" / ("%s.jinja2" % device["hostname"])).open(
            "w", encoding="utf-8"
        ) as f_out:
            f_out.write(str(device_dict))

    export = {
        "aliases": aliases,
        "devices": devices,
        "device-types": device_types,
        "tags": tags,
        "workers": workers,
    }

    # Dump the configuration
    with (dest / "instance.yaml").open("w", encoding="utf-8") as f_out:
        flow_yaml.dump(export, f_out)
    return 0


def handle_maintenance(proxy, options, _):
    print("Maintenance workers:")
    workers = proxy.scheduler.workers.list()
    excluded_workers = []
    if options.exclude:
        excluded_workers = options.exclude.split(",")
    for worker in workers:
        if worker in excluded_workers:
            print("* %s [SKIP]" % worker)
            continue
        print("* %s" % worker)
        proxy.scheduler.workers.update(worker, None, "MAINTENANCE")

    excluded_devices = []
    if excluded_workers:
        for worker in excluded_workers:
            excluded_devices.extend(proxy.scheduler.workers.show(worker)["devices"])

    print("Wait for devices:")
    devices = proxy.scheduler.devices.list()
    for device in devices:
        if device["hostname"] in excluded_devices:
            continue
        print("* %s" % device["hostname"])
        current_job = device["current_job"]
        if current_job is not None:
            print("--> waiting for job %s" % current_job)
            # if --force is passed, cancel the job
            if options.force:
                print("---> canceling")
                proxy.scheduler.jobs.cancel(current_job)
            while proxy.scheduler.jobs.show(current_job)["state"] != "Finished":
                print("---> waiting")
                time.sleep(5)
    return 0


def handle_methods(proxy, options, _):
    if options.sub_sub_sub_command == "help":
        print(proxy.system.methodHelp(options.method))
    elif options.sub_sub_sub_command == "signature":
        print(proxy.system.methodSignature(options.method))
    else:
        # Fallback to "list"
        methods = proxy.system.listMethods()
        for method in methods:
            print(method)
    return 0


def handle_version(proxy, _, __):
    print(proxy.system.version())
    return 0


def handle_whoami(proxy, _, __):
    username = proxy.system.whoami()
    if username is None:
        print("<AnonymousUser>")
    else:
        print(username)
    return 0


def handle_certificate(proxy, _, __):
    print(proxy.system.get_master_certificate())


def handle(proxy, options, config):
    handlers = {
        "active": handle_active,
        "api": handle_api,
        "certificate": handle_certificate,
        "export": handle_export,
        "maintenance": handle_maintenance,
        "methods": handle_methods,
        "version": handle_version,
        "whoami": handle_whoami,
    }
    return handlers[options.sub_sub_command](proxy, options, config)



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys
import xmlrpc.client

import zmq

from lavacli import main


class DummySocket:
    data = []
    url = "tcp://lava.example.com:5500"

    def setsockopt(self, key, value):
        assert key == zmq.SUBSCRIBE  # nosec
        assert value == b""  # nosec

    def connect(self, url):
        assert url == self.url  # nosec

    def recv_multipart(self):
        return self.data.pop(0)


class DummyContext:
    def socket(self, sock_type):
        assert sock_type == zmq.SUB  # nosec
        return DummySocket()


def test_events_listen(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "events", "listen"])
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    DummySocket.data = [
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Idle",
                    "health": "Good",
                }
            ),
        ),
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Running",
                    "health": "Good",
                    "job": "1234",
                }
            ),
        ),
        (
            "v.l.o.testjob",
            "uuid",
            "2018-01-30",
            "lava-health",
            json.dumps(
                {
                    "job": "1234",
                    "device": "bbb-01",
                    "state": "Running",
                    "health": "Unknown",
                    "description": "a nice job",
                }
            ),
        ),
        ("invalid message"),
        (
            "v.l.o.worker",
            "uuid",
            "2018-01-31",
            "admin",
            json.dumps(
                {"hostname": "worker-01", "state": "Active", "health": "Maintenance"}
            ),
        ),
        (
            "v.l.o.event",
            "uuid",
            "2018-01-31",
            "lavaserver",
            json.dumps({"message": "hello from the job", "job": "1245"}),
        ),
    ]
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.get_publisher_event_socket",
                "args": (),
                "ret": "tcp://*:5500",
            },
        ],
    )
    assert main() == 1  # nosec
    readouterr = capsys.readouterr()
    assert (  # nosec
        readouterr[0]
        == """Listening to tcp://lava.example.com:5500
\033[1;30m2018-01-29\033[0m \033[1;37mv.l.o.device\033[0m \033[32mlava-health\033[0m - [bbb-01] <bbb> state=Idle health=Good
\033[1;30m2018-01-29\033[0m \033[1;37mv.l.o.device\033[0m \033[32mlava-health\033[0m - [bbb-01] <bbb> state=Running health=Good for 1234
\033[1;30m2018-01-30\033[0m \033[1;37mv.l.o.testjob\033[0m \033[32mlava-health\033[0m - [1234] <bbb-01> state=Running health=Unknown (a nice job)
\033[1;30m2018-01-31\033[0m \033[1;37mv.l.o.worker\033[0m \033[32madmin\033[0m - [worker-01] state=Active health=Maintenance
\033[1;30m2018-01-31\033[0m \033[1;37mv.l.o.event\033[0m \033[32mlavaserver\033[0m - [1245] message=hello from the job
"""
    )
    assert (  # nosec
        readouterr[1]
        == "Invalid message: invalid message\nUnknown error: pop from empty list\n"
    )


def test_events_listen_config(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "--uri",
            "https://admin:12345@localhost:456/RPC2",
            "events",
            "listen",
        ],
    )
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    monkeypatch.setattr(DummySocket, "url", "tcp://localhost:5501")
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.get_publisher_event_socket",
                "args": (),
                "ret": "tcp://*:5501",
            },
        ],
    )
    assert main() == 1  # nosec
    readouterr = capsys.readouterr()
    assert readouterr[0] == "Listening to tcp://localhost:5501\n"  # nosec
    assert readouterr[1] == "Unknown error: pop from empty list\n"  # nosec


def test_events_listen_config_2(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "events", "listen"])
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    monkeypatch.setattr(DummySocket, "url", "tcp://localhost:789")
    with (tmp_path / "lavacli.yaml").open("w") as f_conf:
        f_conf.write(
            "default:\n  username: admin\n  token: 12345\n  uri: https://localhost:456/RPC2\n  events:\n    uri: tcp://localhost:789\n"
        )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [{"request": "system.version", "args": (), "ret": version}],
    )
    assert main() == 1  # nosec
    readouterr = capsys.readouterr()
    assert readouterr[0] == "Listening to tcp://localhost:789\n"  # nosec
    assert readouterr[1] == "Unknown error: pop from empty list\n"  # nosec


def test_events_listen_filter(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "events", "listen", "--filter", "device"]
    )
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    DummySocket.data = [
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Idle",
                    "health": "Good",
                }
            ),
        ),
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Running",
                    "health": "Good",
                    "job": "1234",
                }
            ),
        ),
        (
            "v.l.o.testjob",
            "uuid",
            "2018-01-30",
            "lava-health",
            json.dumps(
                {
                    "job": "1234",
                    "device": "bbb-01",
                    "state": "Running",
                    "health": "Unknown",
                    "description": "a nice job",
                }
            ),
        ),
        ("invalid message"),
        (
            "v.l.o.worker",
            "uuid",
            "2018-01-31",
            "admin",
            json.dumps(
                {"hostname": "worker-01", "state": "Active", "health": "Maintenance"}
            ),
        ),
        (
            "v.l.o.event",
            "uuid",
            "2018-01-31",
            "lavaserver",
            json.dumps({"message": "hello from the job", "job": "1245"}),
        ),
    ]
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.get_publisher_event_socket",
                "args": (),
                "ret": "tcp://*:5500",
            },
        ],
    )
    assert main() == 1  # nosec
    readouterr = capsys.readouterr()
    assert (  # nosec
        readouterr[0]
        == """Listening to tcp://lava.example.com:5500
\033[1;30m2018-01-29\033[0m \033[1;37mv.l.o.device\033[0m \033[32mlava-health\033[0m - [bbb-01] <bbb> state=Idle health=Good
\033[1;30m2018-01-29\033[0m \033[1;37mv.l.o.device\033[0m \033[32mlava-health\033[0m - [bbb-01] <bbb> state=Running health=Good for 1234
"""
    )
    assert (  # nosec
        readouterr[1]
        == "Invalid message: invalid message\nUnknown error: pop from empty list\n"
    )


def test_events_listen_filter_2(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "events", "listen", "--filter", "device", "--filter", "worker"],
    )
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    DummySocket.data = [
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Idle",
                    "health": "Good",
                }
            ),
        ),
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Running",
                    "health": "Good",
                    "job": "1234",
                }
            ),
        ),
        (
            "v.l.o.testjob",
            "uuid",
            "2018-01-30",
            "lava-health",
            json.dumps(
                {
                    "job": "1234",
                    "device": "bbb-01",
                    "state": "Running",
                    "health": "Unknown",
                    "description": "a nice job",
                }
            ),
        ),
        ("invalid message"),
        (
            "v.l.o.worker",
            "uuid",
            "2018-01-31",
            "admin",
            json.dumps(
                {"hostname": "worker-01", "state": "Active", "health": "Maintenance"}
            ),
        ),
        (
            "v.l.o.event",
            "uuid",
            "2018-01-31",
            "lavaserver",
            json.dumps({"message": "hello from the job", "job": "1245"}),
        ),
    ]
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.get_publisher_event_socket",
                "args": (),
                "ret": "tcp://*:5500",
            },
        ],
    )
    assert main() == 1  # nosec
    readouterr = capsys.readouterr()
    assert (  # nosec
        readouterr[0]
        == """Listening to tcp://lava.example.com:5500
\033[1;30m2018-01-29\033[0m \033[1;37mv.l.o.device\033[0m \033[32mlava-health\033[0m - [bbb-01] <bbb> state=Idle health=Good
\033[1;30m2018-01-29\033[0m \033[1;37mv.l.o.device\033[0m \033[32mlava-health\033[0m - [bbb-01] <bbb> state=Running health=Good for 1234
\033[1;30m2018-01-31\033[0m \033[1;37mv.l.o.worker\033[0m \033[32madmin\033[0m - [worker-01] state=Active health=Maintenance
"""
    )
    assert (  # nosec
        readouterr[1]
        == "Invalid message: invalid message\nUnknown error: pop from empty list\n"
    )


def test_events_wait_device(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "events", "wait", "device", "bbb-02"])
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    DummySocket.data = [
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Idle",
                    "health": "Good",
                }
            ),
        ),
        (
            "v.l.o.testjob",
            "uuid",
            "2018-01-30",
            "lava-health",
            json.dumps(
                {
                    "job": "1234",
                    "device": "bbb-01",
                    "state": "Running",
                    "health": "Unknown",
                    "description": "a nice job",
                }
            ),
        ),
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-02",
                    "device_type": "bbb",
                    "state": "Running",
                    "health": "Good",
                    "job": "1234",
                }
            ),
        ),
    ]
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.get_publisher_event_socket",
                "args": (),
                "ret": "tcp://*:5500",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "Listening to tcp://lava.example.com:5500\n"
    )
    assert DummySocket.data == []  # nosec


def test_events_wait_device_state(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "events", "wait", "device", "bbb-01", "--state", "RUNNING"],
    )
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    DummySocket.data = [
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Idle",
                    "health": "Good",
                }
            ),
        ),
        (
            "v.l.o.testjob",
            "uuid",
            "2018-01-30",
            "lava-health",
            json.dumps(
                {
                    "job": "1234",
                    "device": "bbb-01",
                    "state": "Running",
                    "health": "Unknown",
                    "description": "a nice job",
                }
            ),
        ),
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Running",
                    "health": "Good",
                    "job": "1234",
                }
            ),
        ),
    ]
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.get_publisher_event_socket",
                "args": (),
                "ret": "tcp://*:5500",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "Listening to tcp://lava.example.com:5500\n"
    )
    assert DummySocket.data == []  # nosec


def test_events_wait_device_health(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "events", "wait", "device", "bbb-01", "--health", "MAINTENANCE"],
    )
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    DummySocket.data = [
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Idle",
                    "health": "Good",
                }
            ),
        ),
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Idle",
                    "health": "Maintenance",
                }
            ),
        ),
    ]
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.get_publisher_event_socket",
                "args": (),
                "ret": "tcp://*:5500",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "Listening to tcp://lava.example.com:5500\n"
    )
    assert DummySocket.data == []  # nosec


def test_events_wait_job(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "events", "wait", "job", "1234"])
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    DummySocket.data = [
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Idle",
                    "health": "Good",
                }
            ),
        ),
        (
            "v.l.o.testjob",
            "uuid",
            "2018-01-30",
            "lava-health",
            json.dumps(
                {
                    "job": "1233",
                    "device": "bbb-01",
                    "state": "Running",
                    "health": "Unknown",
                    "description": "a nice job",
                }
            ),
        ),
        (
            "v.l.o.testjob",
            "uuid",
            "2018-01-30",
            "lava-health",
            json.dumps(
                {
                    "job": "1234",
                    "device": "bbb-01",
                    "state": "Running",
                    "health": "Unknown",
                    "description": "a nice job",
                }
            ),
        ),
    ]
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.get_publisher_event_socket",
                "args": (),
                "ret": "tcp://*:5500",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "Listening to tcp://lava.example.com:5500\n"
    )
    assert DummySocket.data == []  # nosec


def test_events_wait_worker(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "events", "wait", "worker", "worker-01"]
    )
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    DummySocket.data = [
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Idle",
                    "health": "Good",
                }
            ),
        ),
        (
            "v.l.o.testjob",
            "uuid",
            "2018-01-30",
            "lava-health",
            json.dumps(
                {
                    "job": "1233",
                    "device": "bbb-01",
                    "state": "Running",
                    "health": "Unknown",
                    "description": "a nice job",
                }
            ),
        ),
        (
            "v.l.o.testjob",
            "uuid",
            "2018-01-30",
            "lava-health",
            json.dumps(
                {
                    "job": "1234",
                    "device": "bbb-01",
                    "state": "Running",
                    "health": "Unknown",
                    "description": "a nice job",
                }
            ),
        ),
        (
            "v.l.o.worker",
            "uuid",
            "2018-01-31",
            "admin",
            json.dumps(
                {"hostname": "worker-02", "state": "Active", "health": "Maintenance"}
            ),
        ),
        (
            "v.l.o.worker",
            "uuid",
            "2018-01-31",
            "admin",
            json.dumps(
                {"hostname": "worker-01", "state": "Active", "health": "Maintenance"}
            ),
        ),
    ]
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.get_publisher_event_socket",
                "args": (),
                "ret": "tcp://*:5500",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "Listening to tcp://lava.example.com:5500\n"
    )
    assert DummySocket.data == []  # nosec


def test_events_wait_worker_invalid_message(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "events", "wait", "worker", "worker-01"]
    )
    monkeypatch.setattr(zmq, "Context", lambda: DummyContext())
    DummySocket.data = [
        (
            "v.l.o.device",
            "uuid",
            "2018-01-29",
            "lava-health",
            json.dumps(
                {
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "state": "Idle",
                    "health": "Good",
                }
            ),
        ),
        (
            "v.l.o.testjob",
            "uuid",
            "2018-01-30",
            "lava-health",
            json.dumps(
                {
                    "job": "1233",
                    "device": "bbb-01",
                    "state": "Running",
                    "health": "Unknown",
                    "description": "a nice job",
                }
            ),
        ),
        ("strange"),
        (
            "v.l.o.worker",
            "uuid",
            "2018-01-31",
            "admin",
            json.dumps(
                {"hostname": "worker-02", "state": "Active", "health": "Maintenance"}
            ),
        ),
        (
            "v.l.o.worker",
            "uuid",
            "2018-01-31",
            "admin",
            json.dumps(
                {"hostname": "worker-01", "state": "Active", "health": "Maintenance"}
            ),
        ),
    ]
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.get_publisher_event_socket",
                "args": (),
                "ret": "tcp://*:5500",
            },
        ],
    )
    assert main() == 0  # nosec
    readouterr = capsys.readouterr()
    assert readouterr[0] == "Listening to tcp://lava.example.com:5500\n"  # nosec
    assert readouterr[1] == "Invalid message: strange\n"  # nosec
    assert DummySocket.data == []  # nosec



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import argparse
import sys
from pathlib import Path

import jinja2
import ruamel.yaml

from lavacli.utils import safe_yaml

from . import jobs


def configure_parser(parser, _):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    # "logs"
    logs = sub.add_parser("logs", help="log helpers")
    logs_sub = logs.add_subparsers(dest="sub_sub_sub_command", help="Sub commands")
    logs_sub.required = True

    # "logs.print"
    logs_print = logs_sub.add_parser("print", help="print log file")
    logs_print.add_argument("filename", type=argparse.FileType("r"), help="log file")
    logs_print.add_argument(
        "--filters",
        default=None,
        type=str,
        help="comma separated list of levels to show",
    )
    logs_print.add_argument(
        "--raw", default=False, action="store_true", help="print raw logs"
    )

    # "templates"
    templates = sub.add_parser("templates", help="template helpers")
    templates_sub = templates.add_subparsers(
        dest="sub_sub_sub_command", help="Sub commands"
    )
    templates_sub.required = True

    # "templates.render"
    # TODO: add an option to be strict about undefined variables
    templates_render = templates_sub.add_parser("render", help="render jinja2 template")
    templates_render.add_argument(
        "filename", type=argparse.FileType("r"), help="template"
    )
    templates_render.add_argument(
        "--path", type=str, action="append", help="templates lookup path"
    )
    templates_render.add_argument(
        "--context", type=str, action="append", default=[], help="context variables"
    )
    templates_render.add_argument(
        "--trim", action="store_true", default=False, help="trim blocks"
    )


def help_string():
    return "utility functions"


def handle_logs(_, options):
    try:
        logs = safe_yaml.load(options.filename.read())
    except ruamel.yaml.YAMLError:
        print("Invalid yaml file")
    else:
        jobs.print_logs(logs, options.raw, options.filters)


def handle_templates(_, options):
    assert options.sub_sub_sub_command == "render"
    # Set the default value here
    # If we give it to argparse, then it's always present even when --path is
    # passed.
    if options.path is None:
        options.path = ["/etc/lava-server/dispatcher-config/device-types/"]
        if Path("/usr/share/lava-server/device-types/").exists():
            options.path.append("/usr/share/lava-server/device-types/")

    env = jinja2.Environment(  # nosec - used to render yaml
        autoescape=False,
        trim_blocks=options.trim,
        loader=jinja2.FileSystemLoader(options.path),
    )
    data = options.filename.read()
    template = env.from_string(data)

    # Build the context
    try:
        context = {k.split("=")[0]: "=".join(k.split("=")[1:]) for k in options.context}
    except Exception:
        print("Invalid context variable", file=sys.stderr)
        print("Should be of the form key=value", file=sys.stderr)
        return 1
    try:
        print(template.render(**context))
        return 0
    except jinja2.TemplateNotFound:
        print("Unable to find the templates", file=sys.stderr)
        return 1


def handle(proxy, options, _):
    handlers = {"logs": handle_logs, "templates": handle_templates}
    return handlers[options.sub_sub_command](proxy, options)



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys
import xmlrpc.client

from lavacli import main


def test_aliases_add_before_2019_05(setup, monkeypatch, capsys):
    version = "2019.04"
    monkeypatch.setattr(sys, "argv", ["lavacli", "aliases", "add", "new_alias"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "scheduler.aliases.add", "args": ("new_alias",), "ret": None},
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_aliases_add(setup, monkeypatch, capsys):
    version = "2019.05"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "aliases", "add", "new_alias", "device-type"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.add",
                "args": ("new_alias", "device-type"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_aliases_delete(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "aliases", "delete", "new_alias"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.delete",
                "args": ("new_alias",),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_aliases_list_empty(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "aliases", "list"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "scheduler.aliases.list", "args": (), "ret": []},
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Aliases:\n"  # nosec


def test_aliases_list(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "aliases", "list"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.list",
                "args": (),
                "ret": ["first-alias", "second-alias"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "Aliases:\n* first-alias\n* second-alias\n"
    )


def test_aliases_list_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "aliases", "list", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.list",
                "args": (),
                "ret": ["first-alias", "second-alias"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == '["first-alias", "second-alias"]\n'  # nosec


def test_aliases_list_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "aliases", "list", "--yaml"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.list",
                "args": (),
                "ret": ["first-alias", "second-alias"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "[first-alias, second-alias]\n"  # nosec


def test_aliases_show_before_2019_05(setup, monkeypatch, capsys):
    version = "2019.04"
    monkeypatch.setattr(sys, "argv", ["lavacli", "aliases", "show", "my_alias"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.show",
                "args": ("my_alias",),
                "ret": {"name": "my_alias", "device_types": ["qemu", "kvm"]},
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == "name        : my_alias\ndevice-types:\n* qemu\n* kvm\n"
    )


def test_aliases_show(setup, monkeypatch, capsys):
    version = "2019.05"
    monkeypatch.setattr(sys, "argv", ["lavacli", "aliases", "show", "my_alias"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.show",
                "args": ("my_alias",),
                "ret": {"name": "my_alias", "device_type": "qemu"},
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "name       : my_alias\ndevice-type: qemu\n"
    )


def test_aliases_show_json_before_2019_05(setup, monkeypatch, capsys):
    version = "2019.04"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "aliases", "show", "my_alias", "--json"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.show",
                "args": ("my_alias",),
                "ret": {"name": "my_alias", "device_types": ["qemu", "kvm"]},
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == {  # nosec
        "name": "my_alias",
        "device_types": ["qemu", "kvm"],
    }


def test_aliases_show_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "aliases", "show", "kvm", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.show",
                "args": ("kvm",),
                "ret": {"name": "kvm", "device_type": "qemu"},
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == {  # nosec
        "name": "kvm",
        "device_type": "qemu",
    }


def test_aliases_show_yaml_before_2019_05(setup, monkeypatch, capsys):
    version = "2019.04"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "aliases", "show", "my_alias", "--yaml"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.show",
                "args": ("my_alias",),
                "ret": {"name": "my_alias", "device_types": ["qemu", "kvm"]},
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "device_types: [qemu, kvm]\nname: my_alias\n"
    )


def test_aliases_show_yaml(setup, monkeypatch, capsys):
    version = "2019.05"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "aliases", "show", "my_alias", "--yaml"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.aliases.show",
                "args": ("my_alias",),
                "ret": {"name": "my_alias", "device_type": "qemu"},
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "{device_type: qemu, name: my_alias}\n"  # nosec



# vim: set ts=4

# Copyright 2023-present Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys

from lavacli.utils import safe_yaml


def configure_parser(parser, version):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    if version < (2023, 3):
        return

    # "add"
    users_add = sub.add_parser("add", help="add a user")
    users_add.add_argument("username", help="username")
    users_add.add_argument("--email", default=None, help="email")
    users_add.add_argument("--first-name", default=None, help="first name")
    users_add.add_argument("--last-name", default=None, help="last name")
    users_add.add_argument(
        "--no-active",
        action="store_false",
        dest="is_active",
        default=True,
        help="no active user",
    )
    users_add.add_argument(
        "--staff",
        action="store_true",
        dest="is_staff",
        default=False,
        help="staff user",
    )
    users_add.add_argument(
        "--superuser",
        action="store_true",
        dest="is_superuser",
        default=False,
        help="superuser",
    )
    users_add.add_argument(
        "--ldap",
        action="store_true",
        dest="ldap",
        help="ldap user",
    )

    # "delete"
    users_del = sub.add_parser("delete", help="delete a user")
    users_del.add_argument("username", help="username")

    # "groups"
    users_groups = sub.add_parser("groups", help="manage user groups")
    groups_sub = users_groups.add_subparsers(
        dest="sub_sub_sub_command", help="Sub commands"
    )
    groups_sub.required = True
    groups_add = groups_sub.add_parser("add", help="add a group")
    groups_add.add_argument("username", help="username")
    groups_add.add_argument("group", help="name of the group")

    groups_del = groups_sub.add_parser("delete", help="remove a group")
    groups_del.add_argument("username", help="username")
    groups_del.add_argument("group", help="name of the group")

    groups_list = groups_sub.add_parser("list", help="list groups for a user")
    groups_list.add_argument("username", help="username")
    out_format = groups_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        help="print as yaml",
    )

    # "perms"
    users_perms = sub.add_parser("perms", help="manage user permissions")
    perms_sub = users_perms.add_subparsers(
        dest="sub_sub_sub_command", help="Sub commands"
    )
    perms_sub.required = True
    perms_add = perms_sub.add_parser("add", help="add a permission")
    perms_add.add_argument("username", help="username")
    perms_add.add_argument("app", help="application")
    perms_add.add_argument("model", help="model")
    perms_add.add_argument("codename", help="codename")

    perms_del = perms_sub.add_parser("delete", help="remove a permssion")
    perms_del.add_argument("username", help="username")
    perms_del.add_argument("app", help="application")
    perms_del.add_argument("model", help="model")
    perms_del.add_argument("codename", help="codename")

    perms_list = perms_sub.add_parser("list", help="list permissions for a user")
    perms_list.add_argument("username", help="username")
    out_format = perms_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        help="print as yaml",
    )

    # "list"
    users_list = sub.add_parser("list", help="list users")
    out_format = users_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        default=None,
        action="store_const",
        const="yaml",
        help="print as yaml",
    )

    # "show"
    users_show = sub.add_parser("show", help="show user details")
    users_show.add_argument("username", help="username")
    out_format = users_show.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        default=None,
        help="print as yaml",
    )

    # "update"
    users_update = sub.add_parser("update", help="update a user")
    users_update.add_argument("username", help="username")
    users_update.add_argument("--email", default=None, help="email")
    users_update.add_argument("--first-name", default=None, help="first name")
    users_update.add_argument("--last-name", default=None, help="last name")
    users_update.add_argument(
        "--active",
        action="store_true",
        dest="is_active",
        default=None,
        help="active user",
    )
    users_update.add_argument(
        "--no-active",
        action="store_false",
        dest="is_active",
        default=None,
        help="no active user",
    )
    users_update.add_argument(
        "--staff",
        action="store_true",
        dest="is_staff",
        default=None,
        help="staff user",
    )
    users_update.add_argument(
        "--no-staff",
        action="store_false",
        dest="is_staff",
        default=None,
        help="no staff user",
    )
    users_update.add_argument(
        "--superuser",
        action="store_true",
        dest="is_superuser",
        default=None,
        help="superuser",
    )
    users_update.add_argument(
        "--no-superuser",
        action="store_false",
        dest="is_superuser",
        default=None,
        help="no superuser",
    )


def help_string():
    return "manage users"


def handle_add(proxy, options, _):
    proxy.auth.users.add(
        options.username,
        options.first_name,
        options.last_name,
        options.email,
        options.is_active,
        options.is_staff,
        options.is_superuser,
        options.ldap,
    )
    return 0


def handle_delete(proxy, options, _):
    proxy.auth.users.delete(options.username)
    return 0


def handle_groups(proxy, options, _):
    if options.sub_sub_sub_command == "add":
        proxy.auth.users.groups.add(options.username, options.group)
    elif options.sub_sub_sub_command == "delete":
        proxy.auth.users.groups.delete(options.username, options.group)
    else:
        groups = proxy.auth.users.groups.list(options.username)
        if options.output_format == "json":
            print(json.dumps(groups))
        elif options.output_format == "yaml":
            safe_yaml.dump(groups, sys.stdout)
        else:
            print("Groups:")
            for group in groups:
                print("* %s" % group)
    return 0


def handle_list(proxy, options, _):
    users = proxy.auth.users.list()
    if options.output_format == "json":
        print(json.dumps(users))
    elif options.output_format == "yaml":
        safe_yaml.dump(users, sys.stdout)
    else:
        print("Users:")
        for u in users:
            msg = "* %s" % u["username"]
            if u["first_name"] or u["last_name"]:
                msg += " (%s %s)" % (u["first_name"], u["last_name"])
            if u["is_staff"]:
                msg += " [staff]"
            if u["is_superuser"]:
                msg += " [superuser]"
            if not u["is_active"]:
                msg += " [INACTIVE]"
            print(msg)
    return 0


def handle_perms(proxy, options, _):
    if options.sub_sub_sub_command == "add":
        proxy.auth.users.perms.add(
            options.username, options.app, options.model, options.codename
        )
    elif options.sub_sub_sub_command == "delete":
        proxy.auth.users.perms.delete(
            options.username, options.app, options.model, options.codename
        )
    else:
        perms = proxy.auth.users.perms.list(options.username)
        if options.output_format == "json":
            print(json.dumps(perms))
        elif options.output_format == "yaml":
            safe_yaml.dump(perms, sys.stdout)
        else:
            print("Permissions:")
            for perm in perms:
                print("* %s.%s.%s" % (perm["app"], perm["model"], perm["codename"]))
    return 0


def handle_show(proxy, options, config):
    user = proxy.auth.users.show(options.username)

    if options.output_format == "json":
        user["date_joined"] = user["date_joined"].value or None
        user["last_login"] = user["last_login"].value or None
        print(json.dumps(user))
    elif options.output_format == "yaml":
        user["date_joined"] = user["date_joined"].value or None
        user["last_login"] = user["last_login"].value or None
        safe_yaml.dump(user, sys.stdout)
    else:
        print("username   : %s (%d)" % (user["username"], user["id"]))
        print("first name : %s" % user["first_name"])
        print("last name  : %s" % user["last_name"])
        print("groups     : %s" % ", ".join(user["groups"]))
        print("permissions: %s" % ", ".join(user["permissions"]))
        print("email      : %s" % user["email"])
        print("active     : %s" % user["is_active"])
        print("staff      : %s" % user["is_staff"])
        print("superuser  : %s" % user["is_superuser"])
        print("date joined: %s" % user["date_joined"])
        print("last login : %s" % user["last_login"])


def handle_update(proxy, options, _):
    proxy.auth.users.update(
        options.username,
        options.first_name,
        options.last_name,
        options.email,
        options.is_active,
        options.is_staff,
        options.is_superuser,
    )
    return 0


def handle(proxy, options, config):
    handlers = {
        "add": handle_add,
        "delete": handle_delete,
        "groups": handle_groups,
        "list": handle_list,
        "perms": handle_perms,
        "show": handle_show,
        "update": handle_update,
    }
    return handlers[options.sub_sub_command](proxy, options, config)



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import sys

from lavacli import main
from lavacli.__about__ import __version__


def test_lavacli_version(setup, monkeypatch, capsys):
    monkeypatch.setattr(sys, "argv", ["lavacli", "--version"])

    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "lavacli %s\n" % __version__  # nosec


def test_lavacli_missing_arguments(setup, monkeypatch, capsys):
    monkeypatch.setattr(sys, "argv", ["lavacli", "--help"])
    assert main() == 0  # nosec
    help_out = capsys.readouterr()[0]

    monkeypatch.setattr(sys, "argv", ["lavacli"])
    assert main() == 1  # nosec
    assert capsys.readouterr()[0] == help_out  # nosec


def test_lavacli_wrong_identity(setup, monkeypatch, capsys):
    monkeypatch.setattr(sys, "argv", ["lavacli", "-i", "bla", "devices", "list"])

    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Unknown identity 'bla'\n"  # nosec



# vim: set ts=4

# Copyright 2022-present Linaro
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

from voluptuous import All, Any, Invalid, Length, Range, Schema
from voluptuous.validators import Email

group_device_permission = Schema(
    {
        "name": All(str, Length(min=1)),
        "group": All(str, Length(min=1)),
    }
)

device_schema = Schema(
    {
        "hostname": All(str, Length(min=1)),
        "device_type": All(str, Length(min=1)),
        "worker": Any(str, None),
        "description": str,
        "tags": list,
        "permissions": [group_device_permission],
        "retire": bool,
    },
)


device_types_schema = Schema(
    Any(
        {
            "name": All(str, Length(min=1)),
            "description": str,
            "health_disabled": bool,
            "health_denominator": Any("hours", "jobs"),
            "health_frequency": All(int, Range(min=1)),
            "aliases": list,
            "display": bool,
            "permissions": [group_device_permission],
        },
        None,
    )
)

group_schema = Schema(Any({"name": str, "permissions": list}, None))

user_schema = Schema(
    Any(
        {
            "username": All(str, Length(min=1)),
            "last_name": All(str, Length(min=1)),
            "first_name": All(str, Length(min=1)),
            "email": Email(),
            "is_superuser": bool,
            "is_staff": bool,
            "is_active": bool,
            "ldap": bool,
            "groups": list,
            "permissions": list,
        },
        None,
    )
)

worker_schema = Schema(
    Any(
        {
            "hostname": str,
            "description": str,
            "job_limit": All(int, Range(min=1)),
            "retire": bool,
        },
        None,
    )
)


def user_group_must_exist(data: dict) -> dict:
    users = data.get("users", {})
    for user in users:
        user_config = users[user]
        if user_config:
            for group in user_config.get("groups", []):
                if group not in data.get("groups", {}):
                    raise Invalid(
                        f"Group must exist",
                        path=["users", user, "groups", group],
                    )

    return data


def device_dt_must_exist(data: dict) -> dict:
    devices = data.get("devices", {})
    for device in devices:
        dt = devices[device].get("device_type")
        if dt and dt not in data.get("device_types", {}):
            raise Invalid(
                f"Device type must exist",
                path=["devices", device, "device_type", dt],
            )

    return data


def device_worker_must_exist(data: dict) -> dict:
    devices = data.get("devices", {})
    for device in devices:
        worker = devices[device].get("worker")
        if worker and worker not in data.get("workers", {}):
            raise Invalid(
                f"Worker must exist",
                path=["devices", device, "worker", worker],
            )

    return data


def device_perms_group_must_exist(data: dict) -> dict:
    devices = data.get("devices", {})
    for device in devices:
        perms = devices[device].get("permissions", [])
        for perm in perms:
            group = perm["group"]
            if group not in data.get("groups", {}):
                raise Invalid(
                    f"Group must exist",
                    path=["devices", device, "permissions", perm],
                )

    return data


def dt_perms_group_must_exist(data: dict) -> dict:
    dts = data.get("device_types", {})
    for dt in dts:
        dt_config = dts[dt]
        if dt_config:
            perms = dts[dt].get("permissions", [])
            for perm in perms:
                group = perm["group"]
                if group not in data.get("groups", {}):
                    raise Invalid(
                        f"Group must exist",
                        path=["devices_types", dt, "permissions", perm],
                    )

    return data


config_schema = Schema(
    All(
        {
            "groups": {str: group_schema},
            "users": {str: user_schema},
            "workers": {str: worker_schema},
            "device_types": {str: device_types_schema},
            "devices": {str: device_schema},
        },
        user_group_must_exist,
        device_dt_must_exist,
        device_worker_must_exist,
        device_perms_group_must_exist,
        dt_perms_group_must_exist,
    )
)



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import re
import xmlrpc.client
from urllib.parse import urlparse

from ruamel.yaml import YAML

VERSION_LATEST = (3000, 1)

# yaml objects

flow_yaml = YAML(typ="safe")
flow_yaml.default_flow_style = False
log_yaml = YAML(typ="safe")
log_yaml.default_flow_style = True
log_yaml.default_style = '"'
log_yaml.width = 10**6
# 'safe' -> SafeLoader/SafeDumper
safe_yaml = YAML(typ="safe")
# 'rt'/None -> RoundTripLoader/RoundTripDumper (default)
rt_yaml = YAML(typ="rt")


def parse_version(version):
    pattern = re.compile(r"(?P<major>20\d{2})\.(?P<minor>\d{1,2})")
    if not isinstance(version, str):
        version = str(version)
    m = pattern.match(version)
    if m is None:
        return VERSION_LATEST
    res = m.groupdict()
    return (int(res["major"]), int(res["minor"]))


def exc2str(exc, url):
    if isinstance(exc, xmlrpc.client.ProtocolError):
        msg = exc.errmsg
        if url is None:
            return msg
        p = urlparse(url)
        if "@" in p.netloc:
            uri = f"{p.scheme}://<USERNAME>:<TOKEN>@{p.netloc.split('@')[-1]}{p.path}"
        else:
            uri = f"{p.scheme}://{p.netloc}{p.path}"
        return msg.replace(url, uri)
    return str(exc)


def fetch_jobs(proxy, state, health, limit, since, start=0):
    job_numbers = []
    batch_size = 100

    while len(job_numbers) < limit:
        remaining = limit - len(job_numbers)
        fetch_count = min(batch_size, remaining)

        try:
            jobs = proxy.scheduler.jobs.list(
                state,
                health,
                start,
                fetch_count,
                since,
                False,
            )
        except xmlrpc.client.Error as exc:
            print(f"Error fetching jobs: {exc}")
            break

        if not jobs:
            break

        job_ids = [str(job["id"]) for job in jobs]
        job_numbers.extend(job_ids)
        start += batch_size

    return job_numbers


def extract_device_name(job_output):
    device_regex = re.compile(r"(?i)^\s*device\s*:\s*(.*)")
    for line in job_output.splitlines():
        match = device_regex.match(line)
        if match:
            return match.group(1).strip()
    return None


def print_error_summary(error_summary, summary=True):
    if summary:
        print("\n=== Short log with error count type of error and devices ===")
    else:
        print("\n=== Unique Errors Found with job-url ===")

    for error, device_jobs in sorted(
        error_summary.items(),
        key=lambda x: sum(len(jobs) for jobs in x[1].values()),
        reverse=True,
    ):
        total_count = sum(len(jobs) for jobs in device_jobs.values())
        if summary:
            devices_str = ", ".join(sorted(device_jobs))
            print(f"[{total_count}] {error} (Devices: {devices_str})")
        else:
            print(f"\n[{total_count}] {error}")
            for device in sorted(device_jobs):
                print(f"  - {device}:")
                for url in sorted(device_jobs[device]):
                    print(f"    * {url}")



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import subprocess

__all__ = [
    "__author__",
    "__author_email__",
    "__description__",
    "__license__",
    "__url__",
    "__version__",
]


def git_describe():
    try:
        # git describe?
        out = subprocess.check_output(["git", "describe"], stderr=subprocess.STDOUT)
        return out.decode("utf-8").rstrip("\n")[1:]
    except Exception:
        return "git"


__author__ = "Rémi Duraffort"
__author_email__ = "remi.duraffort@linaro.org"
__description__ = "LAVA XML-RPC command line interface"
__license__ = "AGPLv3+"
__url__ = "https://gitlab.com/lava/lavacli"
__version__ = "2.3.0"
